from dataclasses import dataclass

from dataclasses_json import dataclass_json

from ..._helpers import extract_bytes, extract_string, extract_uint8


@dataclass_json
@dataclass
class NurReaderInfo:
    serial: str = None
    alt_serial: str = None
    name: str = None
    fcc_id: str = None
    hw_version: str = None
    sw_version: str = None
    dev_build: int = None
    num_gpio: int = None
    num_sensors: int = None
    num_regions: int = None
    num_antennas: int = None
    max_antennas: int = None


def extract_sw_version(payload: bytearray) -> str:
    major = str(payload[0])
    minor = str(payload[1])
    build = chr(payload[2])
    del payload[:4]
    return major + '.' + minor + '.' + build


def parse_get_reader_info_response(payload: bytearray):
    reader_info = NurReaderInfo()
    unknown = extract_bytes(payload, 4)
    reader_info.serial = extract_string(payload)
    reader_info.alt_serial = extract_string(payload)
    reader_info.name = extract_string(payload)
    reader_info.fcc_id = extract_string(payload)
    reader_info.hw_version = extract_string(payload)
    reader_info.sw_version = extract_sw_version(payload)
    reader_info.num_sensors = extract_uint8(payload)
    reader_info.num_regions = extract_uint8(payload)
    reader_info.num_antennas = extract_uint8(payload)
    reader_info.max_antennas = extract_uint8(payload)
    return reader_info



from dataclasses import dataclass

from dataclasses_json import dataclass_json

from ..._helpers import extract_int16, extract_uint8


@dataclass_json
@dataclass
class InventoryResponse:
    tags_found: int = None
    tags_in_memory: int = None
    rounds: int = None
    collisions: int = None
    current_q: int = None


def parse_inventory_response(payload: bytearray):
    inventory_response = InventoryResponse()
    inventory_response.tags_found = extract_int16(payload)
    inventory_response.tags_in_memory = extract_int16(payload)
    inventory_response.rounds = extract_uint8(payload)
    inventory_response.collisions = extract_int16(payload)
    inventory_response.current_q = extract_uint8(payload)
    return inventory_response



from dataclasses import dataclass
from typing import List

from dataclasses_json import dataclass_json

from ..._helpers import extract_uint8, extract_int8, extract_uint16, extract_uint32


@dataclass_json
@dataclass
class NurTagDataMeta:
    rssi: int = None
    scaled_rssi: int = None
    timestamp: int = None
    frequency: float = None
    pc: bytearray = None
    channel: int = None
    antenna_id: int = None
    epc: bytearray = None


def parse_get_id_buffer_meta_response(payload: bytearray) -> List[NurTagDataMeta]:
    tag_list: List[NurTagDataMeta] = []
    while len(payload) > 0:
        block_length = extract_int8(payload)
        nur_tag_data_meta = NurTagDataMeta()
        nur_tag_data_meta.rssi = extract_int8(payload)
        nur_tag_data_meta.scaled_rssi = extract_int8(payload)
        nur_tag_data_meta.timestamp = extract_uint16(payload)
        nur_tag_data_meta.frequency = extract_uint32(payload)
        nur_tag_data_meta.pc = extract_uint16(payload)
        nur_tag_data_meta.channel = extract_uint8(payload)
        nur_tag_data_meta.antenna_id = extract_uint8(payload)
        # epc_length_words = (nur_tag_data_meta.pc & 0b1111100000000000) >> 11
        # epc_length_bytes = epc_length_words*2
        epc_length_bytes = block_length - 12
        nur_tag_data_meta.epc = payload[:epc_length_bytes]
        payload = payload[epc_length_bytes:]
        tag_list.append(nur_tag_data_meta)
    return tag_list



# SPDX-FileCopyrightText: 2024-present Iz2k <ibon@zalbide.com>
#
# SPDX-License-Identifier: MIT
import logging
import struct
from typing import Callable, List

from .protocol import Packet, CommandCode
from ._helpers import to_uint16_bytes, to_uint8_bytes
from .protocol.command.get_device_capabilities import NurDeviceCaps
from .protocol.command.get_id_buffer_meta import NurTagDataMeta
from .protocol.command.get_mode import Mode
from .protocol.command.get_reader_info import NurReaderInfo
from .protocol.command.inventory import InventoryResponse
from .protocol.command.inventory_stream import InventoryStreamNotification
from .protocol.command.module_setup import ModuleSetup, ModuleSetupFlags, populate_module_setup_args
from .protocol.rx_handler import RxHandler
from .transport.serial import SerialPort

logger = logging.getLogger(__name__)


class NurAPY:

    def __init__(self, connection_string=None):
        self.transport = None
        self._rx_handler = RxHandler()
        self.connection_string = connection_string
        if connection_string is not None:
            self.connect(connection_string)

    def set_notification_callback(self, notification_callback: Callable[[InventoryStreamNotification,
                                                                         List[NurTagDataMeta]], None]):
        self._rx_handler.set_notification_callback(notification_callback)

    def connect(self, connection_string=None) -> bool:
        if connection_string:
            self.connection_string = connection_string
        # TODO: Parse connection string to determine transport type
        self.transport = SerialPort(read_callback=self._rx_handler.append_data)
        return self.transport.connect(connection_string)

    def is_connected(self) -> bool:
        if self.transport is None:
            return False
        return self.transport.is_connected()

    def disconnect(self) -> bool:
        if not self.is_connected():
            logger.info('Transport already disconnected.')
            return True
        try:
            self.transport.disconnect()
            logger.info('Transport successfully disconnected.')
            return True
        except Exception as e:
            logger.warning(e)
            return False

    def _execute_command(self, command_packet: Packet):
        if not self.transport.is_connected():
            if self.connection_string is None:
                logger.info('Transport is disconnected.')
                return None
            if not self.transport.connect(self.connection_string):
                logger.info('Transport is disconnected.')
                return None

        logger.info('TX -> ' + command_packet.get_command_code().name)
        self.transport.write(command_packet.bytes())
        try:
            response = self._rx_handler.get_response()
            logger.info('RX <- ' + str(response))
            return response
        except TimeoutError:
            logger.warning('Timeout executing ' + command_packet.get_command_code().name)
            return None

    def ping(self) -> bool:
        packet = Packet(command_code=CommandCode.PING, args=[0x01, 0x00, 0x00, 0x00])
        response = self._execute_command(packet)
        return response

    def reset(self) -> bool:
        packet = Packet(command_code=CommandCode.RESET, args=[])
        response = self._execute_command(packet)
        return response

    def restart(self) -> bool:
        packet = Packet(command_code=CommandCode.RESTART, args=[])
        response = self._execute_command(packet)
        return response

    def get_mode(self) -> Mode:
        packet = Packet(command_code=CommandCode.GET_MODE, args=[])
        response = self._execute_command(packet)
        return response

    def get_reader_info(self) -> NurReaderInfo:
        packet = Packet(command_code=CommandCode.GET_READER_INFO, args=[])
        response = self._execute_command(packet)
        return response

    def get_device_capabilities(self) -> NurDeviceCaps:
        packet = Packet(command_code=CommandCode.GET_DEVICE_CAPABILITIES, args=[])
        response = self._execute_command(packet)
        return response

    def get_module_setup(self, setup_flags: List[ModuleSetupFlags]) -> ModuleSetup:
        combined_module_setup_flags = 0
        for setup_flag in setup_flags:
            combined_module_setup_flags |= setup_flag.value
        packet = Packet(command_code=CommandCode.GET_MODULE_SETUP, args=[struct.pack('I', combined_module_setup_flags)])
        response = self._execute_command(packet)
        return response

    def set_module_setup(self, setup_flags: List[ModuleSetupFlags], module_setup: ModuleSetup) -> ModuleSetup:
        combined_module_setup_flags = 0
        for setup_flag in setup_flags:
            combined_module_setup_flags |= setup_flag.value
        args = populate_module_setup_args(combined_module_setup_flags, module_setup)
        packet = Packet(command_code=CommandCode.GET_MODULE_SETUP, args=args)
        response = self._execute_command(packet)
        return response

    def simple_inventory(self, q=None, session=None, rounds=None) -> InventoryResponse:
        args = []
        if q is not None and session is not None:
            args.append(q)
            args.append(session)
            if rounds is not None:
                args.append(to_uint8_bytes(rounds))
        packet = Packet(command_code=CommandCode.SIMPLE_INVENTORY, args=args)
        response = self._execute_command(packet)
        return response

    def get_id_buffer(self, clear: bool = False):
        command_code = CommandCode.GET_ID_BUFFER
        args = []
        if clear:
            args.append(0x01)
        packet = Packet(command_code=command_code, args=[])
        response = self._execute_command(packet)
        return response

    def get_id_buffer_with_metadata(self, clear: bool = False):
        command_code = CommandCode.GET_ID_BUFFER_META
        args = []
        if clear:
            args.append(0x01)
        packet = Packet(command_code=command_code, args=[])
        response = self._execute_command(packet)
        return response

    def clear_id_buffer(self):
        packet = Packet(command_code=CommandCode.CLEAR_ID_BUFFER, args=[])
        response = self._execute_command(packet)
        return response

    def start_inventory_stream(self) -> bool:
        packet = Packet(command_code=CommandCode.INVENTORY_STREAM, args=[0x00])
        response = self._execute_command(packet)
        return response

    def stop_inventory_stream(self) -> bool:
        packet = Packet(command_code=CommandCode.INVENTORY_STREAM, args=[])
        response = self._execute_command(packet)
        return response



import logging
import time
from threading import Thread

import serial
from serial.serialutil import SerialException

logger = logging.getLogger(__name__)


class SerialPort:
    def __init__(self, port=None, read_callback=None):
        self.read_callback = read_callback
        self.serial = serial.Serial(
            baudrate=115200,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            bytesize=serial.EIGHTBITS
        )
        self._rx_thread = None
        self._rx_thread_run = False
        if port:
            self.connect(port=port)

    def connect(self, port) -> bool:
        if self.serial.isOpen():
            logger.info('SerialPort already connected.')
            return True
        try:
            self.serial.port = port
            self.serial.open()
            logger.info('SerialPort successfully connected.')
            self._rx_thread = Thread(target=self._rx_thread_fxn, daemon=True, name='RxThread')
            self._rx_thread_run = True
            self._rx_thread.start()
            return True
        except Exception as e:
            logger.warning(e)
            return False

    def is_connected(self) -> bool:
        return self.serial.isOpen()

    def disconnect(self) -> bool:
        self._rx_thread_run = False
        if self._rx_thread:
            self._rx_thread.join()
        if not self.is_connected():
            logger.info('SerialPort already disconnected.')
            return True
        try:
            self.serial.close()
            logger.info('SerialPort successfully disconnected.')
            return True
        except Exception as e:
            logger.warning(e)
            return False

    def write(self, data: bytes):
        self.serial.write(data)
        logger.debug('TX >> ' + data.hex(sep=' ').upper())

    def set_read_callback(self, callback):
        self.read_callback = callback

    def _read(self) -> bytes | None:
        try:
            data = self.serial.read_all()
            if len(data) > 0:
                logger.debug('RX << ' + str(data.hex(sep=' ').upper()))
                return data
        except SerialException:
            logger.info('SerialPort disconnected.')
            self.serial.close()

    def _rx_thread_fxn(self):
        while self._rx_thread_run:
            if self.is_connected():
                data = self._read()
                if data is not None:
                    if len(data) > 0:
                        if self.read_callback:
                            self.read_callback(data)
            time.sleep(0.001)



import logging
import time
from typing import List

from src.nurapy import NurAPY, ModuleSetupFlags, ModuleSetup
from src.nurapy.protocol.command.get_id_buffer_meta import NurTagDataMeta
from src.nurapy.protocol.command.inventory_stream import InventoryStreamNotification
from src.nurapy.protocol.command.module_setup import ModuleSetupLinkFreq, ModuleSetupRxDec

logging.basicConfig(level=logging.DEBUG)

reader = NurAPY('COM8')
if not reader.ping():
    logging.error('Could not connect to NUR APY')
    exit()
reader.get_mode()
reader_info = reader.get_reader_info()
device_caps = reader.get_device_capabilities()
setup = reader.get_module_setup(setup_flags=[
    ModuleSetupFlags.LINKFREQ,
    ModuleSetupFlags.RXDEC,
    ModuleSetupFlags.TXLEVEL,
    ModuleSetupFlags.ANTMASKEX,
    ModuleSetupFlags.SELECTEDANT,
    ModuleSetupFlags.ALL
])

new_setup = ModuleSetup()
new_setup.link_freq = ModuleSetupLinkFreq.BLF_256
new_setup.rx_decoding = ModuleSetupRxDec.MILLER_4
new_setup.tx_level = 0
new_setup.antenna_mask = 1
new_setup.selected_antenna = 0
setup = reader.set_module_setup(setup_flags=[
    ModuleSetupFlags.LINKFREQ,
    ModuleSetupFlags.RXDEC,
    ModuleSetupFlags.TXLEVEL,
    ModuleSetupFlags.ANTMASK,
    ModuleSetupFlags.SELECTEDANT
], module_setup=new_setup)

inventory_response = reader.simple_inventory()
if inventory_response.tags_in_memory:
    tags = reader.get_id_buffer_with_metadata(clear=True)
    logging.info(tags)


def my_notification_callback(inventory_stream_notification: InventoryStreamNotification,
                             tags: List[NurTagDataMeta]):
    if inventory_stream_notification.stopped:
        logging.info('Restarting inventory stream')
        reader.start_inventory_stream()
    for tag in tags:
        logging.info(tag)
    reader.clear_id_buffer()


reader.set_notification_callback(my_notification_callback)
reader.start_inventory_stream()
time.sleep(2)
reader.stop_inventory_stream()



from dataclasses import dataclass

from dataclasses_json import dataclass_json

from ..._helpers import extract_int16, extract_uint8
from .get_id_buffer_meta import parse_get_id_buffer_meta_response


@dataclass_json
@dataclass
class InventoryStreamNotification:
    stopped: bool = None
    rounds: int = None
    collisions: int = None
    last_q: int = None


def parse_inventory_stream_notification(payload: bytearray):
    inventory_stream_notification = InventoryStreamNotification()
    inventory_stream_notification.stopped = bool(extract_uint8(payload))
    inventory_stream_notification.rounds = extract_uint8(payload)
    inventory_stream_notification.collisions = extract_int16(payload)
    inventory_stream_notification.last_q = extract_uint8(payload)
    tag_data = parse_get_id_buffer_meta_response(payload)
    return [inventory_stream_notification, tag_data]
